/*
 * Created on Jul 6, 2005
 *
 */
package tacle.core;

import org.eclipse.jdt.core.IMethod;
import org.eclipse.jdt.core.IType;
import java.util.Set;

/**
 * Tacle-specific representation of a call site. Client code should
 * not implement this interface: instances should be obtained through
 * methods <code>getXYZ</code> from <code>CallGraphAnalysis</code>.
 * 
 * Eventually we may add subinterfaces <code>StaticCallSite</code> and
 * possibly <code>ConstructorCallSite</code>. The current version provides
 * only the subinterface <code>InstanceCallSite</code>. 
 */

public interface IInstanceCallSite extends ICallSite {

    /**
     * Returns a set of <code>IType</code> references, representing
     * the possible run-time receiver types. For each returned
     * reference <code>x</code>, it is guaranteed that 
     * <blockquote> <pre>
     * getInitializedITypes().contains(x) == true
     * </pre> </blockquote>
     * Not guarantted to be complete, due to the limitations of
     * <code>IType</code> and <code>IMethod</code>. 
     * Current version does not report anonymous types. See 
     * http://presto.cse.ohio-state.edu/tacle/index.html
     * for more details on the current implementation. 
     */
    Set getReceiverITypes();

    /**
     * Returns an <code>IMethod</code> reference for the method called
     * by the call site for the given run-time receiver type. If such
     * a receiver type is not possible (i.e., it is not an element of
     * <code>getReceiverITypes()</code>, returns <code>null</code>.
     * Not guarantted to be complete, due to the limitations of
     * <code>IType</code> and <code>IMethod</code>. 
     * 
     * The current version does not deal with reflective calls and throw 
     * clauses. Methods declared in anonymous classes are not reported.
     * The body of a method declared in an anonymous class created that
     * is declared in a method may be wrongly reported as belonging
     * to the enclosing method. See 
     * http://presto.cse.ohio-state.edu/tacle/index.html
     * for more details on the current implementation.   
     * 
     * Current version works only with features of Java 1.4.
     */
    IMethod getCalledIMethod(IType receiverType);

}
    
