/*
 * Created on May 20, 2005
 *
 */
package tacle.internal.rta;

import java.util.Set;
import java.util.Map;
import java.util.HashSet;
import java.util.HashMap;
import java.util.Iterator;

//import java.io.PrintWriter;
//import java.io.FileWriter;

import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.IPackageFragment;
import org.eclipse.jdt.core.IType;
import org.eclipse.jdt.core.JavaCore;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.dom.CompilationUnit;
import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.Type;
import org.eclipse.jdt.core.dom.ITypeBinding;
import org.eclipse.jdt.core.dom.IMethodBinding;
import org.eclipse.jdt.core.dom.Modifier;
import org.eclipse.jdt.core.dom.TypeDeclaration;
import org.eclipse.jdt.core.ICompilationUnit;

import tacle.internal.rta.data.ASTMethod;
import tacle.internal.rta.data.ASTType;
import tacle.internal.rta.data.CallSiteInfo;
import tacle.internal.rta.data.ClassInfo;
import tacle.internal.rta.data.IWorklistMethod;
import tacle.internal.rta.data.IWorklistType;
import tacle.internal.rta.data.InfoDictionary;
import tacle.internal.rta.data.MethodInfo;
import tacle.internal.types.StartupClasses;

/**
 * @author Mariana L Barca
 *  
 */
public class RTAMethodWorklist {

	// set of classes that could be instantiated
	// member: IWorklistClass
	private Set instantiatedClasses;

	private Set initializedClasses;

	// methods that are reachable
	// is this the same as alreadyAnalyzed?
	// member IWorklistMethod
	private Set reachableMethods;

	// set of methods to be analyzed
	// member: IWorklistMethod
	// because we need to know the declaring class
	// and parsing the method key seems unreliable
	private Set worklist;

	// pending call sites
	// IWorklistClass -> set of MethodSignatures
	// this may need to hold the calling method, for call graph later
	private Map pendingCallSites;

	// manages the type hierarchy
	private TypeManager typeMgr;

	// HELPER MEMBERS

	// methods that were already analyzed
	// member: IWorklistMethod
	private Set alreadyAnalyzed;

	// set of java projects reachable from this workspace
	private Set javaProjects;

	// private cache for comp units
	private Map iCompilationUnits; // ASTType -> ICompilationUnit

	// constructor
	public RTAMethodWorklist(TypeManager typeMgr, Set rm, Set initializedT,
			Set jProjects) {
		instantiatedClasses = new HashSet();
		reachableMethods = rm;
		worklist = new HashSet();
		pendingCallSites = new HashMap();
		this.typeMgr = typeMgr;
		alreadyAnalyzed = new HashSet();
		javaProjects = jProjects;
		initializedClasses = initializedT;
		iCompilationUnits = new HashMap();

		// fill java projects
		IProject[] projects = ResourcesPlugin.getWorkspace().getRoot()
				.getProjects();
		for (int i = 0; i < projects.length; i++) {
			IProject project = projects[i];
			// check project has a java nature
			try {
				if (project.isAccessible()
						&& project.hasNature(JavaCore.NATURE_ID)) {
					javaProjects.add(JavaCore.create(project));
				}
			} catch (CoreException ce) {
				ce.printStackTrace();
			}
		}

		processStartupClasses();
	}

	// classes loaded at startup added to the type hierarchy
	// classes instantiated at startup added to intantiatedClasses
	// methods executed at startup added to worklist and reachableMethods
	private void processStartupClasses() {
		String[][] sClasses = StartupClasses.getClasses();
		for (int i = 0; i < sClasses.length; i++) {
			String pkgName = sClasses[i][0];
			String clsName = sClasses[i][1];
			int pkgId = InfoDictionary.v().getPackageIdForName(pkgName);
			ClassInfo ci = InfoDictionary.v().getClassInfoForName(clsName,
					pkgId);
			if (!clsName.equals("java.lang.Object"))
				addToTypeHierarchy(ci);
		}

		String[][] iClasses = StartupClasses.getInstanceClasses();
		for (int i = 0; i < iClasses.length; i++) {
			String pkgName = iClasses[i][0];
			String clsName = iClasses[i][1];
			int pkgId = InfoDictionary.v().getPackageIdForName(pkgName);
			ClassInfo ci = InfoDictionary.v().getClassInfoForName(clsName,
					pkgId);
			if (ci != null) {
				if (!clsName.equals("java.lang.Object"))
					addToTypeHierarchy(ci);
				addInstantiatedClass(ci);
			}
		}

		String[][] startupMethods = StartupClasses.getMethods();
		for (int i = 0; i < startupMethods.length; i++) {
			String name = startupMethods[i][4];
			int packageId = InfoDictionary.v().getPackageIdForName(
					startupMethods[i][0]);
			ClassInfo declClass = InfoDictionary.v().getClassInfoForName(
					startupMethods[i][1], packageId);
			IWorklistType[] paramTypes = new IWorklistType[(startupMethods[i].length - 5) / 2];
			for (int j = 0; j < paramTypes.length; j++) {
				packageId = InfoDictionary.v().getPackageIdForName(
						startupMethods[i][5 + j * 2]);
				if (packageId > 0) {
					paramTypes[j] = InfoDictionary.v().getClassInfoForName(
							startupMethods[i][6 + j * 2], packageId);
				} else {
					paramTypes[j] = InfoDictionary.v().getOtherTypeForName(
							startupMethods[i][6 + j * 2]);
				}
			}

			MethodInfo mi = InfoDictionary.v().getMethodInfoByName(name,
					declClass, paramTypes);

			if (mi != null) {
				worklist.add(mi);
				reachableMethods.add(mi);
			} else {
				System.out.println("Lib method not found in the summary! "
						+ declClass + ":" + name);
			}
		}
	}

	// we assume the main method is always in user code
	// and we can build AST for it
	public void buildMainMethod() {
		// run a visitor
		ASTNode mainMethodAST = typeMgr.getMainMethodCU();
		IWorklistType cls = typeMgr.getMainDeclaringClass();
		MainMethodVisitor rtaVisitor = new MainMethodVisitor(this, cls);
		mainMethodAST.accept(rtaVisitor);
		if (rtaVisitor.hasStaticInitializer()) {
			String methodKey = cls.getCompleteName() + "/<clinit>";
			ASTMethod clinit = InfoDictionary.v().getASTMethod(methodKey);
			if (clinit == null) {
				clinit = new ASTMethod(methodKey, "<clinit>", cls,
						InfoDictionary.v().getOtherTypeForName("void"),
						new IWorklistType[0]);
			}
			addMethodAfterAnalysis(clinit);
		}

	}

	// used to add methods only after they were visited
	// necessary for adding the main method and static initializers of
	// user classes (we don't know they exist until we traverse the classes)
	public void addMethodAfterAnalysis(IWorklistMethod m) {
		// main method is not meant to be in the worklist
		reachableMethods.add(m);
		markAsAnalyzed(m);
	}

	public void build() {
		while (!worklist.isEmpty()) {
			IWorklistMethod method = (IWorklistMethod) worklist.iterator()
					.next();
			worklist.remove(method);
			markAsAnalyzed(method);

			//System.out.println("Analyzing method:" + method.getKey());
			if (method instanceof MethodInfo) {
				processLibMethod((MethodInfo) method);
			} else {
				//System.out.println("Analyzing method:" + method.getKey());
				// test, keep
				if (!(method.getDeclaringClass() instanceof ASTType)) {
					throw new RuntimeException(method.getKey()
							+ " has declaring class not of type ASTClass!");
				}

				ICompilationUnit icu = findICU((ASTType) method
						.getDeclaringClass());
				if (icu==null) {
					icu = findSecondaryICU((ASTType) method
							.getDeclaringClass());
				}
				CompilationUnit cu = typeMgr.getCU(icu);
				if (cu != null) {
					String methodName = method.getName();
					if (methodName.equals("<init>")) { // constructor
						IWorklistType[] paramTypes = method.getParameters();
						ConstructorVisitor cVisitor = new ConstructorVisitor(
								this, method.getDeclaringClass(), method);
						cu.accept(cVisitor);
						if (!cVisitor.hasCalledAnotherConstructor()) {
							// must call parent constructor
							IWorklistType parent = cVisitor.getParent();
							String key = cVisitor.getParentDefaultConstrKey();

							if (parent != null && key == null) {
								// default constructor was not found in binding
								// but it must exist nonetheless
								// probably in the summary
								key = parent + "/<init>()";
							}

							// add default constructor of this class to the
							// worklist
							if (parent != null && key != null) {
								addCGEdge(method, null, -1, -1, key, "<init>",
										parent, InfoDictionary.v()
												.getOtherTypeForName("void"),
										new IWorklistType[0], false);
							}
						}
						if (cVisitor.hasFinalizer()) {
							// this method is not called by anybody
							// so there are no call sites added to it
							String fKey = cVisitor.getFinalizerKey();

							//System.out.println("has finalizer");
							addCGEdge(method, null, -1, -1, fKey, "finalize",
									method.getDeclaringClass(), InfoDictionary
											.v().getOtherTypeForName("void"),
									new IWorklistType[0], false);
						}
					} else if (methodName.equals("<clinit>")) { // static init
						StaticInitializerVisitor iVisitor = new StaticInitializerVisitor(
								this, method.getDeclaringClass(), method);
						cu.accept(iVisitor);
						if (iVisitor.hasStaticInitializer()) {
							addMethodAfterAnalysis(method);
						}
					} else { // not a special method
						ASTNode methodAST = cu.findDeclaringNode(method
								.getKey());
						if (methodAST != null) {
							MethodVisitor rtaVisitor = new MethodVisitor(this,
									method);
							methodAST.accept(rtaVisitor);
						} else {
							// this looks like a strange situation
							throw new RuntimeException(
									"Method AST is null in non-null cu!"
											+ method.getKey());
						}
					}
				}
			}
		}
	}

	// equivalent to the visitors, but for a summary-based method
	private void processLibMethod(MethodInfo mi) {
		// get the initialized classes, add them to type hierarchy
		Set inits = mi.getInitializedClasses();
		for (Iterator it = inits.iterator(); it.hasNext();) {
			ClassInfo cls = (ClassInfo) it.next();
			addToTypeHierarchy(cls);
		}

		// get instantiateld classes, add to instantiated
		Set insts = mi.getInstantiatedClasses();
		for (Iterator it = insts.iterator(); it.hasNext();) {
			ClassInfo cls = (ClassInfo) it.next();
			addToTypeHierarchy(cls);
			addInstantiatedClass(cls);
		}

		// get call sites
		Set sites = mi.getCallSites();
		for (Iterator it = sites.iterator(); it.hasNext();) {
			CallSiteInfo csi = (CallSiteInfo) it.next();
			if (csi.isVirtual()) {
				// resolve virtual call
				IWorklistType[] params = csi.getTargetMethod().getParameters();
				//check if it is thread start, if it is then call thread run
				if (csi.getTargetMethod().getName().equals("start")
						&& params.length == 0
						&& csi.getTargetMethod().getReturnType().equals("void")
						&& implementsInterface(csi.getVarClass(),
								"java.lang.Runnable")) {
					this.processVirtualCall(mi, csi.getVarClass(), "run",
							params, -1, -1);
				} else if (csi.getTargetMethod().getName().equals(
						"doPrivileged")
						&& params.length >= 1
						&& csi.getVarClass().getCompleteName().equals(
								"java.security.AccessController")) {
					this.processVirtualCall(mi, params[0], "run",
							new IWorklistType[0], -1, -1);
				} else {
					this.processVirtualCall(mi, csi.getVarClass(), csi
							.getTargetMethod().getName(), params, -1, -1);
				}
			} else {
				// not a virtual call
				this.addToTypeHierarchy(csi.getVarClass());
				MethodInfo tgt = csi.getTargetMethod();
				this.addCGEdge(mi, null, -1, -1, tgt, false);
			}
		}
	}

	// check if this ClassInfo implements interface interf
	// directly or indirectly
	// this may be done more elegantly through type manager
	private boolean implementsInterface(ClassInfo ci, String interf) {
		if (ci.getCompleteName().equals(interf))
			return true;
		ClassInfo[] interfs = ci.getInterfaces();
		for (int i = 0; i < interfs.length; i++) {
			if (implementsInterface(interfs[i], interf))
				return true;
		}
		return false;
	}

	// add class to instantiated classes
	// we need to pass the binding because type hierarchy info is
	// needed when resolving pending call sites
	public void addInstantiatedClass(ITypeBinding itb) {
		IWorklistType cls = InfoDictionary.v().getTypeForBinding(itb);
		if (cls instanceof ClassInfo) {
			addInstantiatedClass(cls, cls);
		} else {
			addInstantiatedClass(itb, cls);
		}
	}

	public void addInstantiatedClass(ClassInfo ci) {
		addInstantiatedClass(ci, ci);
	}

	// when a new instantiated class is added
	// need to verify if there are pending call sites
	// that would benefit from this inst class
	// obj can be ClassInfo or ITypeBinding
	private void addInstantiatedClass(Object obj, IWorklistType cls) {
		instantiatedClasses.add(cls);
		resolvePending(obj, cls);
	}

	// newType can be ClassInfo or ITypeBinding
	public void resolvePending(Object newType, IWorklistType cls) {
		// for all pending call sites...
		if (!pendingCallSites.keySet().isEmpty()) {
			for (Iterator it = pendingCallSites.keySet().iterator(); it
					.hasNext();) {
				IWorklistType type = (IWorklistType) it.next(); // it is
				if (pendingCallSites.get(type) != null) {
					//if newClass is subclass of type
					if (typeMgr.getAllSubtypes(type).contains(cls)) {
						Set callSites = (Set) pendingCallSites.get(type);
						for (Iterator it1 = callSites.iterator(); it1.hasNext();) {
							ResolvedCallSite ms = (ResolvedCallSite) it1.next();
							processRcvClass(null, ms, ms.getPos(), ms
									.getLength(), newType, ms.methodName,
									ms.paramTypes);
						}
					}
				}
			}
		}
	}

	// AST-based
	// this method adds to the initialized classes too
	public void addToTypeHierarchy(ITypeBinding tb, boolean initialized) {
		// add key mapping to InfoDictionary
		IWorklistType cls = InfoDictionary.v().getTypeForBinding(tb);
		if (cls instanceof ClassInfo) {
			addToTypeHierarchy((ClassInfo) cls);
		} else {
			if (initialized) {
				addInitializedClass((ASTType) cls);
			}

			ITypeBinding parent = tb.getSuperclass();
			if (parent != null) {
				IWorklistType wParent = InfoDictionary.v().getTypeForBinding(
						parent);
				typeMgr.addInHierarchy(wParent, cls);
				if (wParent instanceof ClassInfo) {
					addToTypeHierarchy((ClassInfo) wParent);
				} else {
					addToTypeHierarchy(parent, true);
				}
			}
			ITypeBinding[] interf = tb.getInterfaces();
			for (int i = 0; i < interf.length; i++) {
				if (interf[i] != null) {
					IWorklistType wInterf = InfoDictionary.v()
							.getTypeForBinding(interf[i]);
					typeMgr.addInHierarchy(wInterf, cls);
					if (wInterf instanceof ClassInfo) {
						addToTypeHierarchy((ClassInfo) wInterf);
					} else {
						addToTypeHierarchy(interf[i], true);
					}
				}
			}
		}
	}

	// summary info based
	// assume that classInfo has only classInfos in the hierarchy
	// this method adds to the initialized classes too
	public void addToTypeHierarchy(ClassInfo ci) {
		addInitializedClass(ci);
		ClassInfo parent = ci.getSuperClass();
		if (parent != null) {
			typeMgr.addInHierarchy(parent, ci);
			addToTypeHierarchy(parent);
		}
		ClassInfo[] interf = ci.getInterfaces();
		for (int i = 0; i < interf.length; i++) {
			typeMgr.addInHierarchy(interf[i], ci);
			addToTypeHierarchy(interf[i]);
		}
	}

	// call graph not used to store edges yet
	public void addCGEdge(IWorklistMethod callingMethod, ResolvedCallSite rcs,
			int pos, int len, String calledMethodKey, String methodName,
			IWorklistType declClass, IWorklistType retType,
			IWorklistType[] paramTypes, boolean isVirtual) {
		addCGEdge(callingMethod, rcs, pos, len, calledMethodKey, methodName,
				declClass, declClass, retType, paramTypes, isVirtual);
	}

	// secondary types have an eclosing top level type that defines the
	// compilation unit, it is needed in rare cases for the decl class
	public void addCGEdge(IWorklistMethod callingMethod, ResolvedCallSite rcs,
			int pos, int len, String calledMethodKey, String methodName,
			IWorklistType declClass, IWorklistType topLevelClass,
			IWorklistType retType, IWorklistType[] paramTypes, boolean isVirtual) {
		MethodInfo target = InfoDictionary.v().getMethodInfoByName(methodName,
				declClass, paramTypes);
		if (target != null) {
			addCGEdge(callingMethod, rcs, pos, len, target, isVirtual);
			// add finalizers when constructors are called
			// this works only for summary classes
			// for user class, dealing with finalizers is different
			// see hasFinalize in ConstructorVisitor
			if (methodName.equals("<init>")) {
				String fKey = declClass + "/voidfinalize()";
				addCGEdge(callingMethod, rcs, pos, len, fKey, "finalize",
						declClass, topLevelClass, InfoDictionary.v()
								.getOtherTypeForName("void"),
						new IWorklistType[0], false);
			}
		} else { // user code
			if (!(topLevelClass instanceof ASTType)) {
				//				throw new RuntimeException(topLevelClass
				//						+ " not instance of ASTClass!");
				// it may be a lib class searching for a non-existent finalizer
				// just exit
				return;
			}
			ICompilationUnit icu = findICU((ASTType) topLevelClass);
			if (icu == null) {
				// decl class is in user classes but something is special about
				// it, maybe it's secondary class
				// search the compilation unit in more detail
				icu = findSecondaryICU((ASTType) topLevelClass);
			}
			CompilationUnit cu = typeMgr.getCU(icu);
			if (cu != null) {
				ASTMethod astTarget = new ASTMethod(calledMethodKey,
						methodName, declClass, retType, paramTypes);
				addCGEdge(callingMethod, rcs, pos, len, astTarget, isVirtual);
			} else {
				System.out.println(declClass + " no cu " + calledMethodKey);
				//throw new RuntimeException(declClass);
			}
		}
	}

	public void addCGEdge(IWorklistMethod callingMethod, ResolvedCallSite rcs,
			int pos, int len, IWorklistMethod target, boolean isVirtual) {

		if (!alreadyAnalyzed.contains(target)) {
			worklist.add(target);
		}
		reachableMethods.add(target);
		if (rcs == null) {
			rcs = new ResolvedCallSite(this, target.getName(), target
					.getParameters(), pos, len, isVirtual);
			// registering with the calling method is necessary
			// only when a new rcs is created
			if (callingMethod != null)
				callingMethod.addResolvedCallSite(rcs);
		}
		rcs.addTarget(target);
	}

	// in this implementation of the algorithm, all call sites
	// are pending
	public ResolvedCallSite addPendingEdge(IWorklistMethod callingMethod,
			IWorklistType type, String calledMethodName,
			IWorklistType[] paramTypes, int pos, int len) {
		if (!pendingCallSites.keySet().contains(type)
				|| (pendingCallSites.get(type) == null)) {
			Set callSites = new HashSet();
			pendingCallSites.put(type, callSites);
		}
		Set callSites = (Set) pendingCallSites.get(type);
		ResolvedCallSite rcs = new ResolvedCallSite(this, calledMethodName,
				paramTypes, pos, len, true);
		callSites.add(rcs);
		callingMethod.addResolvedCallSite(rcs);
		return rcs;
	}

	// find all known subtypes of vartype
	// call process rcv class for them
	public void processVirtualCall(IWorklistMethod callingMethod,
			IWorklistType varType, String methodName,
			IWorklistType[] paramTypes, int pos, int len) {
		ResolvedCallSite rcs = addPendingEdge(callingMethod, varType,
				methodName, paramTypes, pos, len);
		Set subTypes = typeMgr.getAllSubtypes(varType);

		for (Iterator it = subTypes.iterator(); it.hasNext();) {
			IWorklistType subType = (IWorklistType) it.next();
			if (!subType.getCompleteName().equals("")) { // why this test?
				if (subType instanceof ClassInfo) {
					processRcvClass(callingMethod, rcs, pos, len,
							(ClassInfo) subType, methodName, paramTypes);
				} else {
					// we need the key to retrieve the AST
					String key = InfoDictionary.v().getTypeKey(subType);
					if (key == null)
						key = "";
					ASTNode astNode = findASTNode((ASTType) subType, key);
					if (astNode != null) {
						ITypeBinding tb = null;
						if (astNode instanceof Type) {
							Type t = (Type) astNode;
							tb = t.resolveBinding();
						} else if (astNode instanceof TypeDeclaration) {
							TypeDeclaration td = (TypeDeclaration) astNode;
							tb = td.resolveBinding();
						}
						if (tb != null) {
							if (tb.isClass()
									&& !Modifier.isAbstract(tb
											.getDeclaredModifiers())) {
								// process rcv class tb
								processRcvClass(callingMethod, rcs, pos, len,
										tb, methodName, paramTypes);
							} else {
								//System.out.println(tb.getQualifiedName()
								//	+ " is not concrete class");
							}
						}
					} else { // astNode = null, must be in the lib
						throw new RuntimeException(
								subType
										+ " not lib type but not found in user classes!");
					}
				}
			} else { // subType of type ClassInfo
				processRcvClass(callingMethod, rcs, pos, len, subType,
						methodName, paramTypes);
			}
		} // end for loop
	}

	// for a type, check if it is instantiated set
	// if yes, find runtime target method and add new edge
	// tb can be ITypeBinding or ClassInfo
	private void processRcvClass(IWorklistMethod callingMethod,
			ResolvedCallSite rcs, int pos, int len, Object tb,
			String methodName, IWorklistType[] paramTypes) {
		IWorklistMethod method = null;
		if (tb instanceof ITypeBinding) {
			IWorklistType cls = InfoDictionary.v().getTypeForBinding(
					(ITypeBinding) tb);
			if (instantiatedClasses.contains(cls)) {
				method = getMethodInClass((ITypeBinding) tb, methodName,
						paramTypes);
			}
		} else if (tb instanceof ClassInfo) {
			if (instantiatedClasses.contains((ClassInfo) tb)) {
				method = getMethodInClass((ClassInfo) tb, methodName,
						paramTypes);
			}
		}
		if (method != null) {
			//rcs.addTarget(method);
			if (method instanceof MethodInfo) {
				addCGEdge(callingMethod, rcs, pos, len, (MethodInfo) method,
						true);
			} else {
				//System.out.println("x - ADDING TO RESOLVED: " + method);
				addCGEdge(callingMethod, rcs, pos, len, method.getKey(), method
						.getName(), method.getDeclaringClass(), method
						.getReturnType(), method.getParameters(), true);
			}
		}
	}

	public void markAsAnalyzed(IWorklistMethod method) {
		alreadyAnalyzed.add(method);
	}

	// used for testing and displaying results
	// keep commented code!
	public void printStat() {
		// print the set of instantiated classes
		//		System.out.println("\nInstantiated classes");
		//		for (Iterator it1 = instantiatedClasses.iterator(); it1.hasNext();) {
		//			Object instCls = it1.next();
		//			if (instCls instanceof ClassInfo)
		//				System.out.println("Instantiated: "
		//						+ ((ClassInfo) instCls).getCompleteName());
		//			else
		//				System.out.println("Instantiated: " + instCls);
		//		}
		//		System.out.println("\nAlready analyzed:"
		//				+ alreadyAnalyzed.size());
		System.out.println("\nInstantiated classes:"
				+ instantiatedClasses.size());
		System.out
				.println("\nInitialized classes:" + initializedClasses.size());
		System.out.println("\nReachable methods:" + reachableMethods.size());

		System.out.println("Non-lib reachable methods");
		int nonlibrm = 0;
		for (Iterator it4 = reachableMethods.iterator(); it4.hasNext();) {
			Object obj = it4.next();
			if (obj instanceof ASTMethod) {
				nonlibrm++;
				ASTMethod astm = (ASTMethod) obj;
				//				System.out.println(astm.getDeclaringClass().getCompleteName()
				//						+ ":" + astm.getReturnType().getCompleteName() + " "
				//						+ astm.getName() + "(" + astm.getParameters().length
				//						+ ")");
			}
		}
		System.out.println("Non-lib reachable methods: " + nonlibrm);
		System.out.println("Percentage of non-lib reachable methods: "
				+ (100 * nonlibrm) / (double) reachableMethods.size());

		//		try {
		//			PrintWriter pw = new PrintWriter(new FileWriter("c:\\init"));
		//			for (Iterator it = initializedClasses.iterator(); it.hasNext();) {
		//				//InfoDictionary.v().printStat(pw);
		//				pw.println(((IWorklistType)it.next()).getCompleteName());
		//			}
		//			pw.close();
		//		} catch (java.io.IOException ioe) {
		//
		//		}

		// print type hierarchy
		//		System.out.println("\nType hierarchy");
		//		typeMgr.printHierarchy();

		// print info dictionary
		//		InfoDictionary.v().printClassMap();

	}

	//// private helper methods

	// dispatch
	// get the first ancestor that declares the method
	// get the method from the ancestor
	private IWorklistMethod getMethodInClass(ITypeBinding tb,
			String methodName, IWorklistType[] paramTypes) {
		// loops through the hierarchy of classes
		// until the method is found
		while (tb != null) {
			IMethodBinding[] mbs = tb.getDeclaredMethods();
			for (int i = 0; i < mbs.length; i++) {
				if (mbs[i].getName().equals(methodName)
						&& sameTypes(mbs[i].getParameterTypes(), paramTypes)) {
					IWorklistType[] params = new IWorklistType[paramTypes.length];
					for (int j = 0; j < params.length; j++) {
						params[j] = InfoDictionary.v().getTypeForBinding(
								mbs[i].getParameterTypes()[j]);
					}
					IWorklistType returnType = InfoDictionary.v()
							.getTypeForBinding(mbs[i].getReturnType());
					IWorklistType declCls = InfoDictionary.v()
							.getTypeForBinding(mbs[i].getDeclaringClass());
					return new ASTMethod(mbs[i].getKey(), mbs[i].getName(),
							declCls, returnType, params);
				}
			}
			tb = tb.getSuperclass();
		}
		return null;
	}

	// dispatch
	// get the first ancestor that declares the method
	// get the method from the ancestor
	private IWorklistMethod getMethodInClass(ClassInfo ci, String methodName,
			IWorklistType[] paramTypes) {
		IWorklistMethod result = null;
		// loops through the hierarchy of classes
		// until the method is found
		while (result == null && ci != null) {
			Map methodsByName = InfoDictionary.v().getMethodsByClass(ci);
			if (methodsByName != null) {
				for (Iterator it = methodsByName.keySet().iterator(); it
						.hasNext();) {
					String name = (String) it.next();
					if (name.equals(methodName)) {
						Set methods = (Set) methodsByName.get(name);
						if (methods != null) {
							for (Iterator it1 = methods.iterator(); it1
									.hasNext();) {
								MethodInfo mi = (MethodInfo) it1.next();
								if (sameTypes(mi.getParameters(), paramTypes)) {
									//pw.println(" " +
									// (mi==null?"null":mi.getKey()));
									return mi;
								}
							}
						}
					}
				}
			}
			ci = ci.getSuperClass();
		}
		return result;
	}

	// check if these represent the same arrays of types in
	// different representations
	// b can have members of type ITypeBinding or IWorklistClass
	private boolean sameTypes(Object[] b, IWorklistType[] s) {
		if (b.length != s.length)
			return false;
		else {
			IWorklistType temp = null;
			for (int i = 0; i < b.length; i++) {
				if (b[i] instanceof IWorklistType) {
					temp = (IWorklistType) b[i];
				} else if (b[i] instanceof ITypeBinding) {
					temp = InfoDictionary.v().getTypeForBinding(
							(ITypeBinding) b[i]);
				} else {
					return false;
				}
				if (!temp.equals(s[i]))
					return false;
			}
			return true;
		}
	}

	// finds AST for a class identified by the key
	private ASTNode findASTNode(ASTType cls, String key) {
		CompilationUnit cu = null;
		ICompilationUnit icu = findICU(cls);
		if (icu == null) {
			icu = findSecondaryICU(cls);
		}
		cu = typeMgr.getCU(icu);
		if (cu != null) {
			ASTNode result = cu.findDeclaringNode(cls.getCompleteName());
			if (result == null && !key.equals(""))
				result = cu.findDeclaringNode(key);
			return result;
		}
		return null;
	}

	// get the ICompilationUnit from the name of the type
	private ICompilationUnit findICU(ASTType cls) {
		ICompilationUnit result = (ICompilationUnit) iCompilationUnits.get(cls);
		if (result != null) {
			return result;
		}
//		String simpleName = cls.getSimpleName();
//		for (Iterator it = javaProjects.iterator(); result == null
//				&& it.hasNext();) {
//			IJavaProject ijProj = (IJavaProject) it.next();
//			try {
//				IType iType = ijProj.findType(cls.getPackageName(), simpleName);
//				if (iType != null) {
//					result = iType.getCompilationUnit();
//				}
//			} catch (JavaModelException jme) {
//				jme.printStackTrace();
//			}
//		}
//		if (result != null) {
//			iCompilationUnits.put(cls, result);
//		}
		return result;
	}

	// get the ICompilationUnit from the name of the type
	// when the type is secondary type
	// or otherwise not available by regular find or in cache
	private ICompilationUnit findSecondaryICU(ASTType cls) {
		String[] possTypeName = cls.getAllPossSimpleNames();
		ICompilationUnit result = null;
		for (Iterator it = javaProjects.iterator(); result == null
				&& it.hasNext();) {
			IJavaProject ijProj = (IJavaProject) it.next();
			try {
				IPackageFragment[] chl = ijProj.getPackageFragments();
				IPackageFragment pkg = null;
				for (int i = 0; i < chl.length; i++) {
					if (chl[i].getElementName().equals(cls.getPackageName())) {
						pkg = chl[i];
						break;
					}
				}
				if (pkg != null) {
					ICompilationUnit[] units = pkg.getCompilationUnits();
					for (int i = 0; i < units.length; i++) {
						IType[] types = units[i].getAllTypes();
						for (int j = 0; j < types.length; j++) {
							for (int k = 0; k < possTypeName.length; k++) {
								if (types[j].getTypeQualifiedName().equals(
										possTypeName[k]))
									return units[i];
							}
						}
					}
				}
			} catch (JavaModelException jme) {
				jme.printStackTrace();
			}
		}
		if (result != null) {
			iCompilationUnits.put(cls, result);
		}
		return result;
	}

	// add clinits to the worklist
	// and add the class to the set of initialized classes
	private void addInitializedClass(ClassInfo ci) {
		initializedClasses.add(ci);
		// find clinit for ci
		MethodInfo clinit = InfoDictionary.v().getClinitInfo(ci);
		// add clinit to worklist
		if (clinit != null && !alreadyAnalyzed.contains(clinit)) {
			worklist.add(clinit);
			reachableMethods.add(clinit);
		}
	}

	// add FAKE clinits to the worklist
	// the method key of a clinit is class_name/<clinit>
	// and add the class to the set of initialized classes
	private void addInitializedClass(ASTType cls) {
		// test, keep
		if (cls.getCompleteName().equals("boolean")) {
			throw new RuntimeException("ERROR!");
		}
		initializedClasses.add(cls);
		String methodKey = cls.getCompleteName() + "/<clinit>";
		ASTMethod clinit = InfoDictionary.v().getASTMethod(methodKey);
		if (clinit == null) {
			clinit = new ASTMethod(methodKey, "<clinit>", cls, InfoDictionary
					.v().getOtherTypeForName("void"), new IWorklistType[0]);
		}
		// add clinit to worklist
		if (clinit != null && !alreadyAnalyzed.contains(clinit)) {
			worklist.add(clinit);
			// clinit added to reachable is delayed until after the traversal
			//reachableMethods.add(clinit);
		}
	}

}
