/*
 * Created on Jun 24, 2005
 *
 */
package tacle.internal.rta.data;

import java.util.HashSet;
import java.util.Set;

import tacle.internal.rta.ResolvedCallSite;

/**
 * @author Mariana L Barca
 *
 * Representation of methods from ASTs 
 */
public class ASTMethod implements IWorklistMethod {

	private String key;

	private IWorklistType declaringClass;

	private String name;

	private IWorklistType returnType;

	private IWorklistType[] parameters;

	private Set resolvedCallSites;
	
	// constructor
	public ASTMethod(String k, String n,
			IWorklistType dC, IWorklistType rT, IWorklistType[] ps) {
		key = k;
		name = n;
		declaringClass = dC;
		returnType = rT;
		parameters = ps;
		resolvedCallSites = new HashSet();
	}

	// getters
	public Set getResolvedCallSites() {
		return resolvedCallSites;
	}
	
	public IWorklistType getDeclaringClass() {
		return declaringClass;
	}

	public String getKey() {
		return key;
	}

	public String getName() {
		return name;
	}

	public IWorklistType[] getParameters() {
		return parameters;
	}

	public IWorklistType getReturnType() {
		return returnType;
	}

	// setter
	public void addResolvedCallSite(ResolvedCallSite rcs) {
		resolvedCallSites.add(rcs);
	}
	
	// overriding methods from class Object
	public String toString() {
		String pNames = "";
		for (int i = 0; i < getParameters().length-1; i++) {
			pNames += getParameters()[i] + ",";
		}
		if (getParameters().length>0) {
			pNames += getParameters()[getParameters().length-1];
		}
		return getReturnType().toString() + " " + getName()
				+ "(" + pNames + ") KEY:" + getKey();
	}

	public boolean equals(Object other) {
		if (other instanceof ASTMethod) {
			ASTMethod astM = (ASTMethod) other;
			return this.key.equals(astM.key);
		}
		return false;

	}
	
	public int hashCode() {
		return this.key.hashCode();
	}
}
